#!/usr/bin/env python
#
'Browse the local testage in a tree.'
#
#  LGPL - don't steal, don't sue
#

import glob, inspect, os, re, string, sys, unittest, time, traceback
from unittest import _TextTestResult, _WritelnDecorator
import Tkinter, Pmw
from TreeBox import TreeBox
from Debug import db #, inspectObject
tk = Tkinter
iz = inspect

_slash = '/'
_dot = '.'

#  TODO - multiple folders on command line
#         save and restore the window geometries
#         same module name in different folders (just warn!)
#         also trigger a retest when imported modules change (!)
#         put a clicker to the top of the test case in the error list
#         ensure the source locator also works on sub-folders
#         route that pesky Tkinter error traceback into here
#         change the color of tree entries that failed to orange or red?
#         at syntax error time, the system just emits a burst of 
#                 irrelevancies, none of which are clickable
#         get the db() system to pump info into the list box


class BrowserResult(_TextTestResult):
    def __init__(self, stream):
        _TextTestResult.__init__(self, stream, descriptions = 1, verbosity = 0)
        
    #  style note - don't use magic members. None all members
    #  in a constructor
    
        self.myBar = None
        self.myTotal = None
        self.myErrorList = None
    
    def setUserView(self, bar, lister, total):
        self.myBar = bar
        self.myErrorList = lister
        self.myTotal = total

    def addSuccess(self, test):
        self.addEvent(test)
        
    def addEvent(self, test):
        if self.myBar is None:  return

        self.myBar.setProgressFraction(
                        self.getFraction(), 
                        self.getColor()
                        )

    def addFailure(self, test, err):
        ''' observe errors go to the console AND the list box.
            Each will provide the user different features '''
        
        _TextTestResult.addFailure(self, test, err) 
        self._printErrorList('FAIL', test, err)
        self.addEvent(test)
      
    def _printErrorList(self, flavour, test, err):
        ''' Any similarity between this function and the one in
            unittest.py must be just a shocking coincidence '''

        description = self.getDescription(test)
        complaint = '%s: %s' % (flavour, description)
        self.myErrorList.insert('end', complaint)
            
        for line in apply(traceback.format_exception, err):
            for l in string.split(line,"\n")[:-1]:
                self.myErrorList.insert('end', l)
        return
        
    def getDescription(self, test):
        return test.shortDescription() or str(test)

    def addError(self, test, err):
        _TextTestResult.addFailure(self, test, err)
        self._printErrorList('ERROR', test, err)
        self.addEvent(test)
        
    def getFraction(self):
        fraction = self.testsRun / float(self.myTotal)
        return fraction

    def getColor(self):
        color = 'cyan'
        
        if len(self.errors) > 0:
            color = 'red'
        elif len(self.failures) > 0:
            color = 'orange'
        elif self.testsRun / self.myTotal:
            color = 'green'
        
        return color
              

class BrowserNode:
    'Behavior & interface common to any clickable in the tree'

    def __init__(self):
        self.myChildren = []
        self.stream = _WritelnDecorator(sys.stdout)
 
    def getChild(self, idx):
        if idx >= len(self.myChildren):  return None
        return self.myChildren[idx]

    def getDoc(self):  return iz.getdoc(self.getThing())
    def getSourceLocation(self, extra = 0):  return (None, None) 
    def uniqueness(self):  return ''
    def getTestCount(self):  return self.getThing().countTestCases()

    def growTree(self, aTree, parentKey = ''):
        name = self.getName()
        key = parentKey + _slash + name + self.uniqueness()
        aTree.AddNode(key, name, parentKey, Data = self)
        
        for child in self.myChildren:
            child.growTree(aTree, key)
        
        return

    def getThing(self):
        ''' If we are not a TestSuite or TestCase, build a TestSuite
            to represent what we are '''
            
        aSuite = unittest.TestSuite()
        
        for suite in self.myChildren:
            aSuite.addTest(suite.getThing())
            
        return aSuite

    def _makeResult(self):  return BrowserResult(self.stream)
    def checkTime(self):  return 0  #  non-file things never change their time

    def checkTimes(self):
        anyChanged = self.checkTime()
        
        for child in self.myChildren:
            if child.checkTimes():
                anyChanged = 1
                
        return anyChanged
        
    def callTests(self, result):
        startTime = time.time()
        self.getThing()(result)
        stopTime = time.time()
        timeTaken = float(stopTime - startTime)
        result.printErrors()
        self.stream.writeln(result.separator2)
        run = result.testsRun

        self.stream.writeln("Ran %d test%s in %.3fs" %
                            (run, run == 1 and "" or "s", timeTaken))

        self.stream.writeln()
        
        if not result.wasSuccessful():
            self.stream.write("FAILED (")
            failed, errored = map(len, (result.failures, result.errors))
            
            if failed:
                self.stream.write("failures=%d" % failed)
                
            if errored:
                if failed: self.stream.write(", ")
                self.stream.write("errors=%d" % errored)
                
            self.stream.writeln(")")
        else:
            self.stream.writeln("Okay")

        return


def _createSourceTuple(thing):
    'this wastes all of [0] - screw it!'
    return (iz.getsourcefile(thing), iz.findsource(thing)[1])


def _findFunctionSource(functions, methodName):
    for name, funk in functions:
        if name == methodName:
            return _createSourceTuple(funk)

    
class TreeTestCase(BrowserNode):

    def __init__(self, aCase):
        BrowserNode.__init__(self)
        self.myCase = aCase

    def getDoc(self):
        return self.myCase._TestCase__testMethodDoc

    def getExtraLocation(self, extra = 0):
        sourceFile = iz.getsourcefile(self.getThing().__class__)
        module = getTesteeModule(sourceFile)

        if module:
    
    #  observe we assume all test methods are named after their
    #  testee method prefixed with test_
    
            methodName = self.getName()[len('test_'):]
            functions = iz.getmembers(module, iz.isfunction)
            source = _findFunctionSource(functions, methodName)
            if source is not None:  return source
            classes = iz.getmembers(module, iz.isclass)

            for className, klass in classes:
                functions = iz.getmembers(klass, iz.ismethod)
                source = _findFunctionSource(functions, methodName)
                if source is not None:  return source

        return (None, None)

    def getSourceLocation(self, extra = 0):
        if extra == 1:
            return self.getExtraLocation()

        method = getattr(self.getThing(), self.getName())
        return _createSourceTuple(method)
        
    def getName(self):  return self.myCase._TestCase__testMethodName
    def getThing(self):  return self.myCase
    def getTestCount(self):  return 1


def _traverseTestList(tests):    

    children = []
    
    for test in tests:
        if test.__class__.__name__ == 'TestSuite':
            aSuite = TreeTestSuite(test)
            q = 0
            
            for child in children:
                if child.getName() == aSuite.getName():
                    q += 1
    
            if q > 0:  aSuite.myUniqueness = str(q)
            children.append(aSuite)
        else:
            aCase = TreeTestCase(test)
            children.append(aCase)
    
    return children

    
class TreeTestSuite(BrowserNode):
    def __init__(self, aSuite):
        BrowserNode.__init__(self)
        self.mySuite = aSuite
        assert aSuite.__class__.__name__ == 'TestSuite'
        self.myChildren += _traverseTestList(self.mySuite._tests)
        self.myUniqueness = ''
        
    def getSourceLocation(self, extra = 0):
        if extra == 1:
            '''module = getTesteeModule(self.myFilePath)
                        
            if module:
                db(iz.getfile(module))
                db(iz.getsourcefile(module))
                return (iz.getsourcefile(module), 0)'''

            return (None, None)
            
        return _createSourceTuple(self.getThing().__class__)
        
    def getName(self):  return self.mySuite.__class__.__name__
    def getThing(self):  return self.mySuite
    def uniqueness(self):  return self.myUniqueness
    def getTestCount(self):  return self.mySuite.countTestCases()
                

class TestFile(BrowserNode):
    def __init__(self, filePath):
        BrowserNode.__init__(self)
        ignoreMe, fileName = os.path.split(filePath)
        moduleName = os.path.splitext(fileName)[0]
        self.myModule = _reImport(moduleName)  
        self.myFilePath = filePath        
        self.myModuleTime = os.path.getmtime(self.myFilePath)
        self._findMySuite()
        
    def _findMySuite(self):
        funcs = iz.getmembers(self.myModule, iz.isfunction)
        myTestSuite = None
        
        for name, method in funcs:
            if name == 'suite':
                myTestSuite = method()
                self.myChildren += _traverseTestList(myTestSuite._tests)

        if myTestSuite is None:
            print 'rejecting', self.getName()
            self.myModule = None  #  suicide

    def checkTime(self):
        nuTime = os.path.getmtime(self.myFilePath)
        if self.myModuleTime == nuTime:  return 0
        self.myModuleTime = nuTime
        # TODO yank out the doofus stray class references
        self.myModule = reload(self.myModule)
        self.myChildren = []
        self._findMySuite()
        # TODO test if self.myModule is now None but user double-clicks
        return 1

    def getName(self):
        return os.path.split(iz.getsourcefile(self.myModule))[1]

    def getDoc(self):  return iz.getdoc(self.myModule)

    def getSourceLocation(self, extra = 0):
        if extra:
            module = getTesteeModule(self.myFilePath)
            
            if module:
                db(iz.getfile(module))
                db(iz.getsourcefile(module))
                return (iz.getsourcefile(module), 0)
                
            return (None, None)
            
        return (self.myFilePath, 0)


def _fileNameToModuleName(fileName):
    'TODO  is there a way in Python to avoid this drek??'

    assert fileName[-3:] == '.py'
    if fileName[0] != '.':  fileName = './' + fileName
    fileName = fileName[:-3]
    moduleName = string.split(fileName, '/')[1:]
    moduleName = string.join(moduleName, '.')
    assert moduleName[0] != '.'
    return moduleName


def getTesteeModule(fileName):
    'if your naming conventions differ, edit here'

    # TODO  make sure this crap does not target any "test" in the folder names
    
    assert string.count(fileName, 'test') > 0

    if string.count(fileName, 'test_') > 0:
        testeeFile = string.join(string.split(fileName, 'test_'), '')
    elif string.count(fileName, '_test') > 0:
        testeeFile = string.join(string.split(fileName, '_test_'), '')
    else:
        testeeFile = string.join(string.split(fileName, 'test'), '')
                
    if not os.path.exists(testeeFile):  return None
    pathThing = _fileNameToModuleName(testeeFile)
    module = __import__(pathThing, globals(), locals(), ['*'])
    return module


def _reImport(moduleName):
    module = __import__(moduleName)
    return reload(module)


_syntaxErrorLine = re.compile(r'^  File "(.*)", line ([0-9]*)')
_testFailureLine = re.compile(r'^(.*):([0-9]*):')

def identifyErrorLine(line):
	'finds file info in a standard error line'

	got = _syntaxErrorLine.search(line)

	if got is None:
		got = _testFailureLine.search(line)

		if got is None:
			return None, None

	group = got.groups(0)

	if len(group) != 2:
		return None, None

	file = group[0]
	line = int(group[1])

	return file, line


class FolderNode(BrowserNode):

    def __init__(self, folder):
        BrowserNode.__init__(self)
        self.myFolder = folder
        self.myInitModule = None

        if not (folder in sys.path):
            sys.path.append(folder)
   
    def getDoc(self):
        if self.myInitModule is None:
            initPath = os.path.join(self.myFolder, '__init__.py')
            
            if not os.path.exists(initPath):
                return 'no __init__.py file to read its doc string'
            
            '''if initPath[:2] == './':
                initPath = initPath[2:]
            if initPath[-3:] == '.py':
                initPath = initPath[:-3]'''
                                
            initPath = _fileNameToModuleName(initPath)
            
            # initPath = string.join(string.split(initPath, '/'), '.')
            # initPath = string.join(string.split(initPath, '\\'), '.')
            
            self.myInitModule = __import__(initPath)  
            
        self.myInitModule = reload(self.myInitModule)  #  TODO do we reload enough?
        return iz.getdoc(self.myInitModule)


class TestFolder(FolderNode):
    'This is so you can run all modules in a folder but not the sub folders'
    
    def __init__(self, folder):
        FolderNode.__init__(self, folder)
       
        listage = glob.glob(os.path.join(folder, '*test*.py'))
        listage.sort()  
        
        for filePath in listage:
            aFile = TestFile(filePath)
                
            if aFile.myModule is not None:
                self.myChildren.append(aFile)

    def getName(self):  return _dot
    

class TestFolderName(FolderNode):
    'This is so you can run a folder and all sub-folders in it'
    
    def __init__(self, folder):
        FolderNode.__init__(self, folder)
        self.myName = os.path.split(folder)[1]  #  OAOO SPLIT
        aFolder = TestFolder(folder)
        
        if aFolder.getChild(0) is not None:
            self.myChildren.append(aFolder)

        _collectSubfolders(self.myChildren, folder)
        
    def getName(self):  return self.myName
    

def _collectSubfolders(children, folder):
    listage = glob.glob(os.path.join(folder, '*'))
    listage.sort()
    
    for thing in listage:
        if os.path.isdir(thing):
            init = os.path.join(thing, '__init__.py')

            if os.path.exists(init):
                aFolder = TestFolderName(thing)

                if aFolder.getChild(0) is not None:
                    children.append(aFolder)


class Browser(BrowserNode):
    'Both the owner of the UI elements and the <all> object'
    
    def __init__(self, folder):
        BrowserNode.__init__(self)
        self.myFolder = folder
        self.myChildren.append(TestFolder(folder))
        self.myBar = None
        self.myTree = None
        self.myErrorList = None
        self.myFrame = None
        _collectSubfolders(self.myChildren, folder)       

    def getName(self):  return '<all>'
    def getDoc(self):  return 'Tap <Enter> to run all tests'
    
    def initializeTree(self):
        '''put the keyboard focus on the tree (you'd be surprised
           how many UIs forget this detail), then expand it 2 ranks.'''
                
        self.myTree.focus_set()
        self.expandRoot()
        self.timer()
        
    def expandRoot(self):
        key1 = _slash + self.getName()
        key2 = key1 + _slash + _dot
        self.myTree.Expand(key1)
        self.myTree.Expand(key2)
        self.myTree.Select(key1)

    def reStuffTree(self):        
        '''always reselects >something<, but will only 
           return 1 if the current node did not go away'''

        db(self.reStuffTree)
        aNode = self.myTree._SelectedNode
        key = aNode['Key']
        formerKey = key
        expanded = aNode['Expanded']
        self.myTree.Clear()
        self.growTree(self.myTree)
        self.expandRoot()

        while len(key) > len(self.getName()):
            if self.myTree.Select(key):
                if expanded or key != formerKey:
                    self.myTree.Expand(key)

                return key == formerKey

            db(key)
            key = string.join(string.split(key, _slash)[:-1], _slash)
        
        return 0  #  current node went away
                
    def timer(self):
        if self.checkTimes():
            if self.reStuffTree():
                aNode = self.myTree.GetSelectedNode('Data')
                self.runSelectedTests(aNode)
                self.startTimer()
        else:

    #  observe this only restarts the timer if the
    #  reloads succeeded

            self.startTimer()
        
    def startTimer(self):
        self.myTree.after(500, self.timer)
        
    def browse(self, root):
        root.title(_getCurrentFolder() + " Test Browser")
        self.myFrame = root
        panes = Pmw.PanedWidget(root)
        top = panes.add('top')
        bottom = panes.add('bottom')
        aTreeBox = self._buildTree(top)
        self.myTree = aTreeBox
        _packMax(aTreeBox)
        self.growTree(aTreeBox)

        self.myErrorList = Pmw.ScrolledListBox(
                bottom,
                 dblclickcommand = self.editErrorInList,
                 #       command = self.editErrorInList,
                selectioncommand = self.editErrorInList,
                )
        self.myBar = ProgressBar(bottom)
        self.myBar.pack(fill = 'x')
        _packMax(self.myErrorList)
        _packMax(panes)        

    def editErrorInList(self, args = None):
        sel = self.myErrorList.getcurselection()
        if len(sel) != 1:  return
        errorLine = sel[0]
        file, line = identifyErrorLine(errorLine)
        if file is None or line is None:  return
        file = _completePath(file)
        self.editSource(file, line)

    def treeNodeClick(self, arg):
        node = self._treeData(arg)
        doc = node.getDoc()
        self.myBar.setMessage(doc, 'white')
    
    def runSelectedItem(self, arg):  
        node = self._treeData(arg)      
        if node is None:  return
        self.runSelectedTests(node)
        
   #  The timer only starts at launch, and after you <Enter>
   #  or double-click a node, and if the timer has successfully run.
   #  If the timer experienced an error reloading modules, it stays
   #  off until the next time you manually test (or restart, of course)        
        
        self.startTimer()
        
    def runSelectedTests(self, node):
        self.myErrorList.clear()
        result = node._makeResult()
        testCount = node.getTestCount()
        result.setUserView(self.myBar, self.myErrorList, testCount)
        node.callTests(result)
                
    def treeContextMenu(self, arg):
        node = self._treeData(arg)
        thing = node.getThing()
        rmenu = tk.Menu(self.myFrame, tearoff = 0)
        file, line = node.getSourceLocation()

        def addEditCommand(rmenu, file, line, concept):
            if file is None or line is None:  return
            file = _completePath(file)
            line += 1  #  some editors are 1-based

            def command(self = self, file=file, line = line):
                self.editSource(file, line)

            rmenu.add_command(
                label = "Edit %s `%s' at %i" % (concept, file, line),
                command = command
                )
        addEditCommand(rmenu, file, line, 'test')
        file, line = node.getSourceLocation(1)
        addEditCommand(rmenu, file, line, 'code')
        rmenu.tk_popup(arg.x_root, arg.y_root)

    def editSource(self, file, line):
        slug = 'vim -g --servername TDD --remote +%i %s &'

        '''TODO get kate working
        # kate is zero based
        line -= 1
        slug = "kate --line %i %s &"
        '''
        
        cmd = slug % (line, file)
        print cmd
        os.system(cmd)
        
    def _treeData(self, arg):
        tree = arg.widget
        if tree._SelectedNode is None:  return None
        return tree.GetSelectedNode('Data')
               
    def _buildTree(self, parent):
        tree = TreeBox(
                        parent,
                        visibleicons = tk.NO, # YES,
                        bg = '#eeeeee',
                        linecolor = 'blue',
                        selcolor = 'lightgreen',
                        signcolor = 'red',
                        font = 'helvetica 14 bold',
                        )
        
        tree.focus_set()
        tree.bind('<<NodeClick-1>>', self.treeNodeClick)
        tree.bind('<<Select>>', self.treeNodeClick)
        tree.bind('<<Return>>', self.runSelectedItem)
        tree.bind('<<NodeClick-2>>', self.treeContextMenu)
        tree.bind('<<NodeClick-3>>', self.treeContextMenu)
        return tree


def _packMax(widget):
    widget.pack(expand = 1, fill = 'both')


def _getFolder():
    if len(sys.argv) == 1:
        folder = '.'
    elif len(sys.argv) != 2:
        print 'Syntax: ' + sys.argv[0] + " <folder>"
        sys.exit(1)
    else:
        folder = sys.argv[1]

    return folder


def _completePath(file):
    return os.path.normcase(os.path.abspath(file))


class ProgressBar(tk.Frame):
    """A simple progress bar that shows a percentage progress in
    the given colour."""

    def __init__(self, *args, **kwargs):
        apply(tk.Frame.__init__, (self,) + args, kwargs)

        self.myCanvas = tk.Canvas(
            self,
            height = 32,
#           width = _width,
            background = 'white',
            borderwidth = 3,
            )
        self.myCanvas.pack(fill=tk.X, expand=1)
        self.rect = self.text = None
        # self.myCanvas.bind('<Configure>', self.paint)
        self.setProgressFraction(0.0, color = 'white')
        self.text = None

    def setProgressFraction(self, fraction, color):
        self.fraction = fraction
        self.color = color
        self.paint()
        self.myCanvas.update_idletasks()

    def paint(self, *args):
        percentString = "%3.0f%%" % (100.0 * self.fraction)
        self.setMessage(percentString)

    def setMessage(self, message, color = None):
        if color is not None:
            self.color = color

        totalWidth = self.myCanvas.winfo_width()
        width = int(self.fraction * float(totalWidth))
        height = self.myCanvas.winfo_height()

        if self.rect is not None:
            self.myCanvas.delete(self.rect)

        if self.text is not None:
            self.myCanvas.delete(self.text)

        self.rect = self.myCanvas.create_rectangle(0, 0, width, height,
                                                 fill=self.color)

        self.text = self.myCanvas.create_text(
                        totalWidth / 2,
                        height / 2,
                        anchor = tk.CENTER,
                        text = message
                )
        return


def _getCurrentFolder():
    dirname = os.getcwd()
    ignoreMe, dirname = os.path.split(dirname)
    return dirname


if __name__ == '__main__':
    root = Pmw.initialise()
    aBrowser = Browser(_getFolder())
    aBrowser.browse(root)
    root.after_idle(aBrowser.initializeTree)
    tk.mainloop()
