#!/usr/bin/env python
#
#  LGPL - don't steal, don't sue
#
'tests the Browser test runner'

import unittest, os, inspect, time, string
from browser import Browser
from browser import identifyErrorLine
from mock import Mock 
from Debug import db, inspectObject, tolerance
iz = inspect


class BrowserTestCase(unittest.TestCase):
    """Tests the TestBrowser class"""
    
    def setUp(self):
        self.tearDown()

    def test_getChild(self):
        aBrowser = _makeCensoredBrowser()
        self.assertEquals(aBrowser.getName(), '<all>')
        aChild = aBrowser.getChild(0)
        self.assertEquals(aChild.getName(), '.')
        aChild = aChild.getChild(0)
        
    #  observe things are generally in alpha order        
        
        self.assertEquals(aChild.getName(), 'alltests.py')
        aChild = aChild.getChild(0)
        self.assertEquals(aChild.getName(), 'TestSuite')
        aChild = aChild.getChild(0)
        self.assertEquals(aChild.getName(), 'TestSuite')
        aChild = aChild.getChild(0)
        self.assertEquals(aChild.getName(), 'testAppend')
        aChild = aBrowser.getChild(1)
        self.assertEquals(aChild, None)

    def test_growTree(self):
        aBrowser = _makeCensoredBrowser()
        aTreeBox = Mock()
        aBrowser.growTree(aTreeBox)
        calls = aTreeBox.getAllCalls()
        
        for call in calls:
            self.assertEquals(call.getName(), 'AddNode')
            self.assertEquals(len(call.params), 3)
            self.assertEquals(len(call.kwparams), 1)
            snip = repr(call.kwparams['Data'])[:9]
            self.assertEquals(snip, '<browser.')
            
        x  = 0;  self.assertEquals(calls[x].params, ('/<all>', '<all>', ''))
        x += 1;  self.assertEquals(calls[x].params, ('/<all>/.', '.', '/<all>'))
        x += 1;  self.assertEquals(calls[x].params, ('/<all>/./alltests.py', 'alltests.py', '/<all>/.'))
        x += 1;  self.assertEquals(calls[x].params, ('/<all>/./alltests.py/TestSuite', 'TestSuite', '/<all>/./alltests.py'))
        x += 1;  self.assertEquals(calls[x].params, ('/<all>/./alltests.py/TestSuite/TestSuite', 'TestSuite', '/<all>/./alltests.py/TestSuite'))
        x += 1;  self.assertEquals(calls[x].params, ('/<all>/./alltests.py/TestSuite/TestSuite/testAppend', 'testAppend', '/<all>/./alltests.py/TestSuite/TestSuite'))
        x += 1;  self.assertEquals(calls[x].params, ('/<all>/./alltests.py/TestSuite/TestSuite/testCount', 'testCount', '/<all>/./alltests.py/TestSuite/TestSuite'))
    
    def test_callTests(self):
        aBrowser = _makeCensoredBrowser()
        aTestCase = aBrowser.getChild(0).getChild(0).getChild(0).getChild(0).getChild(0)
        self.assertEquals(aTestCase.getName(), 'testAppend')
        self.assertEquals(aTestCase.getTestCount(), 1)
        result = aTestCase._makeResult()
        aTestCase.callTests(result)
        self.assertEquals(result.testsRun, 1)
    
        aTestSuite = aBrowser.getChild(0).getChild(0).getChild(0).getChild(0)
        self.assertEquals(aTestSuite.getName(), 'TestSuite')
        self.assertEquals(aTestSuite.getTestCount(), 4)
        result = aTestSuite._makeResult()       
        aTestSuite.callTests(result)
        self.assertEquals(result.testsRun, 4)
        
        aFolder = aBrowser.getChild(0)
        self.assertEquals(aFolder.__class__.__name__, 'TestFolder')
        result = aFolder._makeResult()
        self.assertEquals(aFolder.getName(), '.')
           
        self.assertEquals(aFolder.getTestCount(), 416)
        aFolder.callTests(result)
        self.assertEquals(result.testsRun, 416)
        aFile = aBrowser.getChild(0).getChild(0)
        self.assertEquals(aFile.__class__.__name__, 'TestFile')
        self.assertEquals(aFile.getTestCount(), 8)
        self.assertEquals(aBrowser.getTestCount(), 416)
        
    def test_GreenBar(self):
        aBrowser = _makeCensoredBrowser()
        aFolder = aBrowser.getChild(0)        
        bar, lister = _mockProgress(aFolder)
        calls = bar.getAllCalls()
        
        for call in calls:
            self.assertEquals(call.getName(), 'setProgressFraction')
            
        for call in calls[:-1]:
            self.assertEquals(len(call.params), 2)
            fraction = call.getParam(0)
            self.assert_(fraction >= 0.0 and fraction <= 100.0)
            color = call.getParam(1)
            self.assertEquals(color, 'cyan')

        self._testEndOfBar(calls, 'green')
        self.assertEquals(len(calls), 416)
        
    def _testEndOfBar(self, calls, endColor):
        call = calls[-1]
        self.assertEquals(len(call.params), 2)
        fraction = call.getParam(0)
        self.assert_(fraction >= 0.9999 and fraction <= 1.001)
        self.assert_(tolerance(fraction, 1.0))
        color = call.getParam(1)
        self.assertEquals(color, endColor)

    def test_OrangeBar(
            self,
            problem = 'self.assert_(0)',
            color = 'orange'
            ):
        file(_test_orangeHerring, 'w') \
            .write(_instaTest % problem)
            
        aBrowser = Browser('.')
        aFile = aBrowser.getChild(0).getChild(4)
        fileName = aFile.getName()
        compare = _test_orangeHerring
        
        if fileName[0:2] == './':
            compare = './' + compare

        self.assertEquals(fileName, compare)
        bar, lister = _mockProgress(aFile)
        calls = bar.getAllCalls()
        self._testEndOfBar(calls, color)
        self.assertEquals(len(calls), 1)
        self.tearDown()
        calls = lister.getAllCalls()
        self.assertNotEquals(calls, [])

        for call in calls:
            self.assertEquals(call.getName(), 'insert')
            
        self.assertEquals(calls[0].params, ('end', 'FAIL: testFaylur (test_orangeHerring.RedHerringTestCase)'))
        self.assertEquals(calls[1].params, ('end', 'Traceback (most recent call last):'))

        either = ('end', '  File "test_orangeHerring.py", line 5, in testFaylur')
        ore = ('end', '  File "./test_orangeHerring.py", line 5, in testFaylur')

        self.assert_(calls[2].params == either or calls[2].params == ore)

        self.assertEquals(calls[3].params, ('end', '    self.assert_(0)'))
        self.assertEquals(calls[4].params, ('end', '  File "/usr/lib/python2.2/unittest.py", line 262, in failUnless'))
        self.assertEquals(calls[5].params, ('end', '    if not expr: raise self.failureException, msg'))
        self.assertEquals(calls[6].params, ('end', 'AssertionError'))

    def teest_RedBar(self):
        'TODO does anyone know how to create an error?'
        self.test_OrangeBar('x = 1 / 0', 'red')

    def test_subFolders(self):
        self.tearDown()
        os.mkdir(_subFolder)
        
        file(_subFile, 'w') \
            .write(_instaTest % 'pass')

        aBrowser = Browser('.')
        aFolder = aBrowser.getChild(1)
        
    #  we decline to observe files without a __init__.py
        
        self.assertEquals(aFolder, None)
        
        file(_subInit, 'w') \
            .write('#stub')
        
        aBrowser = Browser('.')
        aFolderName = aBrowser.getChild(1)
        self.assertEquals(aFolderName.getName(), _subFolder)
        aFolder = aFolderName.getChild(0)
        aFile = aFolder.getChild(0)
        self.assertEquals(aFile.getName(), _testBullfrog)

    #  uncomment this to browse the scratch folders
#        os.system('python ./browser.py')

    def test_setUp(self):
        'run a test that only passes if setUp works'
        
        instaTest = '''
import unittest, sys
x = 0
class RedHerringTestCase(unittest.TestCase):
    def setUp(self):
        global x
        x = 1
    def testFaylur(self):
        self.assertEquals(x, 1)
''' + _instaSuite

        os.mkdir(_subFolder)
        file(_subFile, 'w').write(instaTest)
        
    # This tests that Browser obeys folders other than '.'
        
        aBrowser = Browser(_subFolder)
        aFolder = aBrowser.getChild(0)
        result = aFolder._makeResult()
        self.assertEquals(aFolder.getName(), '.')
        self.assertEquals(aFolder.getChild(0).getName(), _testBullfrog)
        self.assertEquals(aFolder.getTestCount(), 1)
        aFolder.callTests(result)
        self.assertEquals(result.testsRun, 1)

    def test_tearDown(self):
        'run a tearDown and inspect it for side-effects'

        instaTest = '''
import unittest, sys
x = 0
class RedHerringTestCase(unittest.TestCase):
    def testFaylur(self):
        pass
    def tearDown(self):
        self.instaMember = 1
''' + _instaSuite

        os.mkdir(_subFolder)
        file(_subFile, 'w').write(instaTest)
        
    # This tests that Browser obeys folders other than '.'
        
        aBrowser = Browser(_subFolder)
        aFolder = aBrowser.getChild(0)
        result = aFolder._makeResult()
        self.assertEquals(aFolder.getTestCount(), 1)
        aFolder.callTests(result)
        self.assertEquals(result.testsRun, 1)
        aRedHerring = aFolder.getChild(0).getChild(0).getChild(0).getThing()
        self.assertEquals(aRedHerring.instaMember, 1)

    # was that sick or what? ;-)

    def test_getDoc(self):  #  also probes getSourceLocation
        'my doc'
        
        try:
            raise 'yo'
        except:
            f, file, lineno, fn_name, lines, n = iz.trace()[-1]
            lineno -= 7
            
    #    lineno = 246  #  yes if you edit above here you must change this
    
        aBrowser = Browser('.')
        idx = 0
        thisTest = None
        aFolder = aBrowser.getChild(0)

    #  a folder's doc string is thet doc from its __init__.py file, if any

        self.assertEquals(aFolder.getDoc(), 'illogical')
        self.assertEquals(aFolder.getSourceLocation(), (None, None))

    #  note nobody needs iterators except this test, so far

        while aFolder.getChild(idx) is not None and thisTest is None:
            if aFolder.getChild(idx).getName() == 'test_browser.py':
                aFile = aFolder.getChild(idx)
                self.assertEquals(aFile.getDoc(), 'tests the Browser test runner')
                self.assertEquals(aFile.getSourceLocation(), (aFile.myFilePath, 0))
                aSuite = aFile.getChild(0)
                self.assertEquals(aSuite.getSourceLocation(), ('/usr/lib/python2.2/unittest.py', 306))
                self.assertEquals(aSuite.getSourceLocation(1), (None, None))
                idx2 = 0

                while aSuite.getChild(idx2) is not None and thisTest is None:
                    aTest = aSuite.getChild(idx2)

                    if aTest.getName() == 'test_getDoc':
                        source = aTest.getSourceLocation()
                        either = ('test_browser.py', lineno)
                        ore = ('./test_browser.py', lineno)
                        self.assert_(source == either or source == ore)
                        thisTest = aTest
            
                    idx2 += 1

            idx += 1

        self.assertNotEquals(thisTest, None)
        doc = thisTest.getDoc()
        self.assertEquals(doc, 'my doc')
                
    def test_checkTime(self):
        file(_test_orangeHerring, 'w') \
            .write(_instaTest % 'pass')
            
        aBrowser = Browser('.')
        aFile = aBrowser.getChild(0).getChild(4)
        wazModule = aFile.myModule
        got = aFile.checkTime()
        self.assert_(not got)
        self.assertEquals(wazModule, aFile.myModule)

    #  now add a test, and change the time

        file(_test_orangeHerring, 'w') \
            .write(_instaTest % 'pass\n    def test_zone(self):\n        pass')
        
        touch(_test_orangeHerring, time.time() - 1000000)
        got = aFile.checkTime()
        self.assert_(got)
#        self.assertNotEquals(wazModule, aFile.myModule)
        self.assertEquals(iz.getfile(wazModule), iz.getfile(aFile.myModule))
        aSuite = aFile.getChild(0)
        aTest = aSuite.getChild(0)
        bTest = aSuite.getChild(1)
        
        self.assert_('test_zone' in [aTest.getName(), bTest.getName()])
        
    #  now peek at the time again, and show checkTime has
    #  a side-effect
    
        got = aFile.checkTime()
        self.assert_(not got)

    def test_identifyErrorLine(self):
		example = '''Traceback (most recent call last):
  File "./testrunner.py", line 16, in ?
	got = runMe()
  File "./OTestRunner.py", line 1196, in runMe
	return t.run()
  File "./OTestRunner.py", line 1128, in run
	self.runGui()
  File "./OTestRunner.py", line 919, in runGui
	suiteList.stuffTree()
  File "./OTestRunner.py", line 734, in stuffTree
	aFolder = TestFolder('.')
  File "./OTestRunner.py", line 220, in __init__
	self.collectFiles()
  File "./OTestRunner.py", line 250, in collectFiles
	aModule = TestModule(f)
  File "./OTestRunner.py", line 336, in __init__
	self.loadModule()
  File "./OTestRunner.py", line 377, in loadModule
	aClass = TestClass(aKey, klass, self)
  File "./OTestRunner.py", line 461, in __init__
	aMethod = TestMethod(self, aKey)
  File "./OTestRunner.py", line 523, in __init__'''

		lines = string.split(example, '\n')

		lineNumbers = [
			16, 1196, 1128, 919, 734,
			220, 250, 336, 377, 461, 523
			]

		for idx in range(0, len(lines), 2):
			file, line = identifyErrorLine(lines[idx])
			self.assertEquals(file, None)
			self.assertEquals(line, None)
			file, line = identifyErrorLine(lines[idx+1])

			if idx == 0:
				self.assertEquals(file, './testrunner.py')
			else:
				self.assertEquals(file, './OTestRunner.py')

			self.assertEquals(line, lineNumbers[idx / 2])

		sample = './OTestRunner_.py:721: TEST(0) Failed.'
		file, line = identifyErrorLine(sample)
		self.assertEquals(file, './OTestRunner_.py')
		self.assertEquals(line, 721)
        
    def tearDown(self):
        for killit, how in (
                            (_test_orangeHerring, os.remove),
                            (_subFile      , os.remove),
                            (_subInit      , os.remove),
                            (_subInit + 'c', os.remove),
                            (_subModule    , os.remove),
                            (_subFolder    , os.rmdir ),
                           ):
            if os.path.exists(killit):
                how(killit)


_subFolder = 'kozmik'
_testBullfrog = 'testBullfrog.py'
_subFile = os.path.join(_subFolder, _testBullfrog)
_subModule = _subFile + 'c'
_subInit = os.path.join(_subFolder, '__init__.py')
    
_instaSuite = '''def suite():
    suite = unittest.makeSuite(RedHerringTestCase)
    return unittest.TestSuite((suite,))
'''

_instaTest = '''
import unittest, sys
class RedHerringTestCase(unittest.TestCase):
    def testFaylur(self):
        %s
''' + _instaSuite

def _mockProgress(aNode):
    result = aNode._makeResult()
    bar = Mock()      
    lister = Mock ()  
    result.setUserView(bar, lister, aNode.getTestCount())
    aNode.callTests(result)
    return bar, lister

            
_test_orangeHerring = 'test_orangeHerring.py'


def _makeCensoredBrowser():
    aBrowser = Browser('.')
    aFolder = aBrowser.getChild(0)        
    assert aFolder.myChildren[3].getName() == 'test_browser.py'
    del aFolder.myChildren[3]  #  prevent infinite recursion ;-)
                               #  also prevent subsequent numbers from
                               #  changing at add test time
    return aBrowser


def touch(filename, new_time = time.time()) :
    os.utime(filename, (new_time, new_time))
            

def suite():
    suite1 = unittest.makeSuite(BrowserTestCase)
    return unittest.TestSuite((suite1,))


if __name__ == '__main__':
    # When this module is executed from the command-line, run all its tests
    unittest.main()

